import h5py
import numpy as np
import tensorflow as tf
import math


def load_dataset(path_train_dataset, path_test_dataset, X_train_column, Y_train_column, X_test_column, Y_test_column, classes_list):

    train_dataset = h5py.File(str(path_train_dataset), "r")
    X_train_orig = np.array(train_dataset[str(X_train_column)][:])  # your train set features
    Y_train_orig = np.array(train_dataset[str(Y_train_column)][:])  # your train set labels

    test_dataset = h5py.File(str(path_test_dataset), "r")
    X_test_orig = np.array(test_dataset[str(X_test_column)][:])  # your test set features
    Y_test_orig = np.array(test_dataset[str(Y_test_column)][:])  # your test set labels

    classes = np.array(classes_list)  # the list of classes

    Y_train_orig = Y_train_orig.reshape((1, Y_train_orig.shape[0]))
    Y_test_orig = Y_test_orig.reshape((1, Y_test_orig.shape[0]))

    return X_train_orig, Y_train_orig, X_test_orig, Y_test_orig, classes


def flatten(X_train_orig, Y_train_orig, X_test_orig, Y_test_orig, classes):

    # Flatten the training and test images
    X_train_flatten = X_train_orig.reshape(X_train_orig.shape[0], -1).T
    X_test_flatten = X_test_orig.reshape(X_test_orig.shape[0], -1).T

    # Normalize image vectors
    X_train = X_train_flatten/255.
    X_test = X_test_flatten/255.

    # Convert training and test labels to one hot matrices
    Y_train = np.eye(len(classes))[Y_train_orig.reshape(-1)].T
    Y_test = np.eye(len(classes))[Y_test_orig.reshape(-1)].T

    return X_train, Y_train, X_test, Y_test


def create_placeholders(n_x, n_y):
    """
    Creates the placeholders for the tensorflow session.

    Arguments:
    n_x -- scalar, size of an image vector (num_px * num_px = 64 * 64 * 3 = 12288)
    n_y -- scalar, number of classes (from 0 to 5, so -> 6)

    Returns:
    X -- placeholder for the data input, of shape [n_x, None] and dtype "tf.float32"
    Y -- placeholder for the input labels, of shape [n_y, None] and dtype "tf.float32"

    Tips:
    - You will use None because it let's us be flexible on the number of examples you will for the placeholders.
      In fact, the number of examples during test/train is different.
    """

    X = tf.placeholder(tf.float32, shape=(n_x, None), name="X")
    Y = tf.placeholder(tf.float32, shape=(n_y, None), name="Y")

    return X, Y


def initialize_parameters(layers_list, seed=1):
    """
    Initializes parameters to build a neural network with tensorflow.

    Returns:
    parameters -- a dictionary of tensors containing the weights and biases
    """

    W = []
    b = []
    parameters = {}

    for l in range(1, len(layers_list) + 1):

        W.append(tf.get_variable("W" + str(l), [layers_list[l], layers_list[l-1], initializer=tf.contrib.layers.xavier_initializer(seed)))

        b.append(tf.get_variable("b" + str(l),
                 [layers_list[l], 1], initializer=tf.zeros_initializer()))

        parameters["W" + str(l)]= W[l]
        parameters["b" + str(l)]= b[l]

    return parameters


def forward_propagation(X, parameters):
    """
    Implements the forward propagation for the model: LINEAR -> RELU -> LINEAR -> RELU -> LINEAR -> SOFTMAX

    Arguments:
    X -- input dataset placeholder, of shape (input size, number of examples)
    parameters -- python dictionary containing your parameters "W1", "b1", "W2", "b2", "W3", "b3"
                  the shapes are given in initialize_parameters

    Returns:
    Z_final_layer -- the output of the last LINEAR unit
    """

    Z= []
    A= []

    # Retrieve the parameters from the dictionary "parameters"
    for l in range(1, len(layers) + 1):

        Z.append(tf.add(tf.matmul(parameters["W" + str(l)], X), parameters["b" + str(l)]))
        A.append(tf.nn.relu(Z[l-1]))

    Z_final_layer= Z[-1]

    return Z_final_layer


def compute_cost(Z_final_layer, Y):
    """
    Computes the cost

    Arguments:
    Z_final_layer -- output of forward propagation (output of the last LINEAR unit)
    Y -- "true" labels vector placeholder, same shape as Z_final_layer

    Returns:
    cost - Tensor of the cost function
    """

    # to fit the tensorflow requirement for tf.nn.softmax_cross_entropy_with_logits(...,...)
    logits= tf.transpose(Z_final_layer)
    labels= tf.transpose(Y)

    cost= tf.reduce_mean(tf.nn.softmax_cross_entropy_with_logits(logits = logits, labels = labels))

    return cost


def random_mini_batches(X, Y, mini_batch_size=32, seed=0):
    """
    Creates a list of random minibatches from (X, Y)

    Arguments:
    X -- input data, of shape (input size, number of examples)
    Y -- true "label" vector
    mini_batch_size - size of the mini-batches, integer
    seed -- this is only for the purpose of grading, so that you're "random minibatches are the same as ours.

    Returns:
    mini_batches -- list of synchronous (mini_batch_X, mini_batch_Y)
    """

    m= X.shape[1]
    mini_batches= []
    np.random.seed(seed)

    # Step 1: Shuffle (X, Y)
    permutation= list(np.random.permutation(m))
    shuffled_X= X[:, permutation]
    shuffled_Y = Y[:, permutation].reshape((Y.shape[0], m))

    # Step 2: Partition (shuffled_X, shuffled_Y). Minus the end case.
    num_complete_minibatches= math.floor(m/mini_batch_size) # number of mini batches of size mini_batch_size in your partitionning
    for k in range(0, num_complete_minibatches):
        mini_batch_X = shuffled_X[:, k * mini_batch_size: k * mini_batch_size + mini_batch_size]
        mini_batch_Y = shuffled_Y[:, k * mini_batch_size: k * mini_batch_size + mini_batch_size]
        mini_batch= (mini_batch_X, mini_batch_Y)
        mini_batches.append(mini_batch)

    # Handling the end case (last mini-batch < mini_batch_size)
    if m % mini_batch_size != 0:
        mini_batch_X = shuffled_X[:, num_complete_minibatches * mini_batch_size: m]
        mini_batch_Y = shuffled_Y[:, num_complete_minibatches * mini_batch_size: m]
        mini_batch= (mini_batch_X, mini_batch_Y)
        mini_batches.append(mini_batch)

    return mini_batches



import math
import numpy as np
import h5py
import matplotlib.pyplot as plt
import tensorflow as tf
from tensorflow.python.framework import ops
from utils import load_dataset, flatten, create_placeholders,
initialize_parameters, forward_propagation, compute_cost, random_mini_batches


class Model():

    def __init__(self, layers_list, learning_rate=0.001, n_epochs=10,
                 minibatch_size=32):

        self.layers_list = layers_list
        self.learning_rate = learning_rate
        self.n_epochs = n_epochs
        self.minibatch_size = minibatch_size

    def model(self, path_train_dataset, path_test_dataset, X_train_column,
              Y_train_column, X_test_column, Y_test_column, classes_list,
              optimizer_algo='adam', print_cost=True):

        # load the datasets
        X_train_orig, Y_train_orig, X_test_orig, Y_test_orig, classes = load_dataset(
            path_train_dataset, path_test_dataset, X_train_column,
            Y_train_column, X_test_column, Y_test_column, classes_list)

        # pre-processing
        X_train, Y_train, X_test, Y_test = flatten(X_train_orig, Y_train_orig,
                                                   X_test_orig, Y_test_orig, classes)

        # to be able to rerun the model without overwriting tf variables
        ops.reset_default_graph()
        # (n_x: input size, m : number of examples in the train set)
        (n_x, m) = X_train.shape
        n_y = Y_train.shape[0]        # n_y : output size
        costs = []                    # To keep track of the cost

        # Create Placeholders of shape (n_x, n_y)
        X, Y = create_placeholders(n_x, n_y)

        # Initialize parameters
        parameters = initialize_parameters(self.layers_list, seed=1)

        # Forward propagation: Build for-propagation in the tensorflow graph
        Z_final_layer = forward_propagation(X, parameters)

        # Cost function: Add cost function to tensorflow graph
        cost = compute_cost(Z_final_layer, Y)

        # Backpropagation: Define the tensorflow optimizer. Use AdamOptimizer
        if optimizer_algo == 'gradient_descent':
            optimizer = tf.train.GradientDescentOptimizer(
                learning_rate=self.learning_rate).minimize(cost)

        elif optimizer_algo == 'momentum':
            optimizer = tf.train.MomentumOptimizer(learning_rate=self.learning_rate).minimize(cost)

        elif optimizer_algo == 'adam':
            optimizer = tf.train.AdamOptimizer(learning_rate=self.learning_rate).minimize(cost)

        # Initialize all the variables
        init = tf.global_variables_initializer()

        # Start the session to compute the tensorflow graph
        with tf.Session() as sess:

            # Run the initialization
            sess.run(init)

            # Do the training loop
            for epoch in range(self.n_epochs):

                epoch_cost = 0.                       # Defines a cost related to an epoch
                # number of minibatches of size minibatch_size in the train set
                num_minibatches = int(m / minibatch_size)
                seed = seed + 1
                minibatches = random_mini_batches(X_train, Y_train, self.minibatch_size, seed)

                for minibatch in minibatches:

                    # Select a minibatch
                    (minibatch_X, minibatch_Y) = minibatch
                    _, minibatch_cost = sess.run([optimizer, cost], feed_dict={
                        X: minibatch_X, Y: minibatch_Y})

                    epoch_cost += minibatch_cost / minibatch_size

                # Print the cost every epoch
                if print_cost == True and epoch % 100 == 0:
                    print("Cost after epoch %i: %f" % (epoch, epoch_cost))

                if print_cost == True and epoch % 5 == 0:
                    costs.append(epoch_cost)

            # lets save the parameters in a variable
            parameters = sess.run(parameters)
            print("Parameters have been trained!")

            # stores quantities useful for later
            quantities = {"X": X, "Y": Y, "Z_final_layer": Z_final_layer,
                          "X_train": X_train, "Y_train": Y_train,
                          "X_test": X_test, "Y_test": Y_test}

        return quantities, costs, parameters

    def plot_costs(self, costs):

        # plot the cost
        plt.plot(np.squeeze(costs))
        plt.ylabel('cost')
        plt.xlabel('iterations (per fives)')
        plt.title("Learning rate =" + str(self.learning_rate))

        plt.show()

    def correct_predictions(self, quantities):

        Y = quantities["Y"]
        Z_final_layer = quantities["Z_final_layer"]

        # Calculate the correct predictions
        correct_predictions = tf.equal(tf.argmax(Z_final_layer), tf.argmax(Y))

        return correct_predictions

    def train_accuracy(self, correct_predictions, quantities):

        X = quantities["X"]
        Y = quantities["Y"]
        X_train = quantities["X_train"]
        Y_train = quantities["Y_train"]

        # Calculate accuracy on the train set
        accuracy = tf.reduce_mean(tf.cast(correct_predictions, "float"))

        train_accuracy = accuracy.eval({X: X_train, Y: Y_train})

        return train_accuracy

    def test_accuracy(self, correct_predictions, quantities):

        X = quantities["X"]
        Y = quantities["Y"]
        X_test = quantities["X_test"]
        Y_test = quantities["Y_test"]

        # Calculate accuracy on the test set
        accuracy = tf.reduce_mean(tf.cast(correct_predictions, "float"))

        test_accuracy = accuracy.eval({X: X_test, Y: Y_test})

        return test_accuracy



from setuptools import setup

setup(name='neural_networks_tfw1',
      version='0.1',
      description='Implementing Neural Networks with Tensorflow',
      packages=['neural_networks_tfw1'],
      author='Dilay Fidan Ercelik',
      author_email='dilay.ercelik@gmail.com',
      zip_safe=False)
